Page
Library
Module
Module type
Parameter
Class
Class type
Tyre.CharsetSourceSets of characters support more operations than regular regexps, as you can diff them, so they have a specific type that allows these operations.
To convert to a regular Tyre.t, use charset or rep_charset
A set of characters.
range of characters ordered according to their code. Include both bounds.
In general, matches latin1 characters, thats is the ocaml Stdlib.char type.
The exact characters matched are not documented in Re, the documentation bellow was written using the source: https://ocaml.orange/p/re/latest/doc/src/re/cset.ml.html .
lower is lowercase latin1 letter.
Includes range 'a' 'z', char 'µ', range '\223' '\246' = set "ßàáâãäåæçèéêëìíîïðñòóôõö" and range '\248' '\255' = set "øùúûüýþÿ"
latin1 ponctuation.
set "!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~\160¡¢£¤¥¦¨©«¬\173®¯°±²³´¶·¸¹»¼½¾¿×÷"upper is latin1 uppercase letter. This includes ascii uppercase letters, that is range 'A' 'Z', but also the ranges range '\192' '\214' = set "ÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖ" and range '\216' '\222' = set "ØÙÚÛÜÝÞ"