You can search for identifiers within the package.
in-package search v0.2.0
tyre
Tyre.Infix
val (<||>) : ('e, 'a) t -> ('e, 'b) t -> ('e, ('a, 'b) Either.t) t
t <||> t' is either t t'.
t <||> t'
either t t'
val (<|>) : (_, 'a) t -> (_, 'a) t -> 'a pattern
t <|> t' is alt t t'. It is not compatible with eval, use either instead if you need to call eval.
t <|> t'
alt t t'
eval
either
val (<&>) : ('e, 'a) t -> ('e, 'b) t -> ('e, 'a * 'b) t
t <&> t' is seq t t'.
t <&> t'
seq t t'
val (*>) : (_, _) t -> ('e, 'a) t -> ('e, 'a) t
ti *> t is prefix ti t.
ti *> t
prefix ti t
val (<*) : ('e, 'a) t -> (_, _) t -> ('e, 'a) t
t <* ti is suffix t ti.
t <* ti
suffix t ti
val (<*>) : ('e, 'a -> 'b) t -> ('e, 'a) t -> 'b pattern
(<*>) is app
(<*>)
app
val (<$>) : ('a -> 'b) -> ('c, 'a) t -> 'b pattern
(<$>) is map
(<$>)
map
val (let+) : ('e, 'a) t -> ('a -> 'b) -> 'b pattern
let+ x = y in z is map (fun x -> z) y.
let+ x = y in z
map (fun x -> z) y
val (and+) : ('e, 'a) t -> ('e, 'b) t -> ('e, 'a * 'b) t
(and+) x y is seq x y.
(and+) x y
seq x y
Be warned that this is not an applicative functor: let+ x = t1 and+ y = t2 in z is not the same as let+ y = t2 and+ x = t1 in z.
let+ x = t1 and+ y = t2 in z
let+ y = t2 and+ x = t1 in z